"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || (function () {
    var ownKeys = function(o) {
        ownKeys = Object.getOwnPropertyNames || function (o) {
            var ar = [];
            for (var k in o) if (Object.prototype.hasOwnProperty.call(o, k)) ar[ar.length] = k;
            return ar;
        };
        return ownKeys(o);
    };
    return function (mod) {
        if (mod && mod.__esModule) return mod;
        var result = {};
        if (mod != null) for (var k = ownKeys(mod), i = 0; i < k.length; i++) if (k[i] !== "default") __createBinding(result, mod, k[i]);
        __setModuleDefault(result, mod);
        return result;
    };
})();
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Inferable = exports.log = void 0;
const debug_1 = __importDefault(require("debug"));
const path_1 = __importDefault(require("path"));
const zod_1 = require("zod");
const create_client_1 = require("./create-client");
const errors_1 = require("./errors");
const links = __importStar(require("./links"));
const machine_id_1 = require("./machine-id");
const polling_1 = require("./polling");
const workflow_1 = require("./workflows/workflow");
const l1m_1 = __importDefault(require("l1m"));
// Custom json formatter
debug_1.default.formatters.J = (json) => {
    return JSON.stringify(json, null, 2);
};
exports.log = (0, debug_1.default)("inferable:client");
/**
 * The Inferable client. This is the main entry point for using Inferable.
 *
 * ```ts
 * // create a new Inferable instance
 * const client = new Inferable({
 *  apiSecret: "API_SECRET",
 * });
 *
 *
 * // Register a tool
 * client.tools.register("hello", z.object({name: z.string()}), async ({name}: {name: string}) => {
 *  return `Hello ${name}`;
 * })
 *
 * await client.tools.listen();
 *
 * // stop the service on shutdown
 * process.on("beforeExit", async () => {
 *   await myService.stop();
 * });
 *
 * ```
 */
class Inferable {
    static getVersion() {
        // eslint-disable-next-line @typescript-eslint/no-var-requires, @typescript-eslint/no-require-imports
        return require(path_1.default.join(__dirname, "..", "package.json")).version;
    }
    /**
     * Initializes a new Inferable instance.
     * @param apiSecret The API Secret for your Inferable cluster. If not provided, it will be read from the `INFERABLE_API_SECRET` environment variable.
     * @param options Additional options for the Inferable client.
     * @param options.endpoint The endpoint for the Inferable cluster. Defaults to https://api.inferable.ai.
     *
     * @example
     * ```ts
     * // Basic usage
     * const client = new Inferable({
     *  apiSecret: "API_SECRET",
     * });
     *
     * // OR
     *
     * process.env.INFERABLE_API_SECRET = "API_SECRET";
     * const client = new Inferable();
     * ```
     */
    constructor(options) {
        this.pollingAgents = [];
        this.toolsRegistry = {};
        this.tools = {
            /**
             * Registers a tool with Inferable.
             * @param input The tool definition.
             * @example
             * ```ts
             * const client = new Inferable({apiSecret: "API_SECRET"});
             *
             * client.tools.register("hello", z.object({name: z.string()}), async ({name}: {name: string}) => {
             *   return `Hello ${name}`;
             * });
             *
             * // start the service
             * await client.tools.listen();
             *
             * // stop the service on shutdown
             * process.on("beforeExit", async () => {
             *   await client.tools.stop();
             * });
             * ```
             */
            register: (input) => {
                this.registerTool({
                    name: input.name,
                    description: input.description,
                    func: input.func,
                    // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    inputSchema: input.schema?.input ?? zod_1.z.object({}).passthrough(),
                    config: input.config,
                });
            },
            listen: async () => {
                if (this.pollingAgents.length > 0) {
                    throw new errors_1.InferableError("Tools already listening");
                }
                // TODO: Create one polling agent per 10 tools
                const agent = new polling_1.PollingAgent({
                    endpoint: this.endpoint,
                    machineId: this.machineId,
                    apiSecret: this.apiSecret,
                    clusterId: await this.getClusterId(),
                    tools: Object.values(this.toolsRegistry),
                });
                this.pollingAgents.push(agent);
                await agent.start();
            },
            unlisten: async () => {
                Promise.all(this.pollingAgents.map((agent) => agent.stop()));
            },
        };
        this.llm = {
            structured: async ({ input, schema, instructions, }, options) => {
                return new l1m_1.default({
                    baseUrl: `${this.endpoint}/clusters/${await this.getClusterId()}/l1m`,
                    additionalHeaders: {
                        Authorization: `Bearer ${this.apiSecret}`,
                    },
                    provider: {
                        key: "",
                        model: "claude-3-5-sonnet",
                        url: "",
                    },
                }).structured({ input, schema, instructions }, options);
            },
        };
        this.workflows = {
            helpers: workflow_1.helpers,
            create: ({ name, description, inputSchema, config, }) => {
                return new workflow_1.Workflow({
                    name,
                    inferable: this,
                    description,
                    config,
                    getClusterId: async () => await this.getClusterId(),
                    client: this.client,
                    inputSchema,
                    endpoint: this.endpoint,
                    machineId: this.machineId,
                    apiSecret: this.apiSecret,
                });
            },
            trigger: async (name, input) => {
                const clusterId = await this.getClusterId();
                const result = await this.client.createWorkflowExecution({
                    params: {
                        clusterId,
                        workflowName: name,
                    },
                    body: input,
                });
                if (result.status !== 201) {
                    throw new errors_1.InferableAPIError("Failed to create workflow execution", result);
                }
            },
        };
        if (options?.apiSecret && process.env.INFERABLE_API_SECRET) {
            (0, exports.log)("API Secret was provided as an option and environment variable. Constructor argument will be used.");
        }
        const apiSecret = options?.apiSecret || process.env.INFERABLE_API_SECRET;
        if (!apiSecret) {
            throw new errors_1.InferableError(`No API Secret provided. Please see ${links.DOCS_AUTH}`);
        }
        if (!apiSecret.startsWith("sk_")) {
            throw new errors_1.InferableError(`Invalid API Secret. Please see ${links.DOCS_AUTH}`);
        }
        this.apiSecret = apiSecret;
        this.endpoint =
            options?.endpoint ||
                process.env.INFERABLE_API_ENDPOINT ||
                "https://api.inferable.ai";
        this.machineId = options?.machineId || (0, machine_id_1.machineId)();
        this.client = (0, create_client_1.createApiClient)({
            baseUrl: this.endpoint,
            machineId: this.machineId,
            apiSecret: this.apiSecret,
        });
    }
    registerTool({ name, func, inputSchema, config, description, }) {
        if (this.toolsRegistry[name]) {
            throw new errors_1.InferableError(`Tool name '${name}' is already registered.`);
        }
        const registration = {
            name,
            func,
            schema: {
                input: inputSchema,
            },
            config,
            description,
        };
        const existing = this.pollingAgents.length > 0;
        if (existing) {
            throw new errors_1.InferableError(`Tools must be registered before starting the listener. Please see ${links.DOCS_FUNCTIONS}`);
        }
        if (typeof registration.func !== "function") {
            throw new errors_1.InferableError(`func must be a function. Please see ${links.DOCS_FUNCTIONS}`);
        }
        (0, exports.log)(`Registering tool`, {
            name: registration.name,
        });
        this.toolsRegistry[registration.name] = registration;
    }
    async getClusterId() {
        if (!this.clusterId) {
            // Call register machine without any services to test API key and get clusterId
            const registerResult = await (0, polling_1.registerMachine)(this.client);
            this.clusterId = registerResult.clusterId;
        }
        return this.clusterId;
    }
}
exports.Inferable = Inferable;
//# sourceMappingURL=Inferable.js.map