"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.registerMachine = exports.PollingAgent = exports.log = void 0;
const debug_1 = __importDefault(require("debug"));
const zod_1 = require("zod");
const create_client_1 = require("./create-client");
const errors_1 = require("./errors");
const serialize_error_1 = require("./serialize-error");
const execute_fn_1 = require("./execute-fn");
const util_1 = require("./util");
const zod_to_json_schema_1 = __importDefault(require("zod-to-json-schema"));
const DEFAULT_RETRY_AFTER_SECONDS = 10;
exports.log = (0, debug_1.default)("inferable:client:polling-agent");
class PollingAgent {
    constructor(options) {
        this.polling = false;
        this.tools = [];
        this.retryAfter = DEFAULT_RETRY_AFTER_SECONDS;
        this.client = (0, create_client_1.createApiClient)({
            baseUrl: options.endpoint,
            machineId: options.machineId,
            apiSecret: options.apiSecret,
        });
        this.tools = options.tools;
        this.clusterId = options.clusterId;
    }
    async start() {
        (0, exports.log)("Starting polling agent");
        await (0, exports.registerMachine)(this.client, this.tools);
        // Purposefully not awaited
        this.runLoop();
    }
    async stop() {
        (0, exports.log)("Stopping polling agent");
        this.polling = false;
    }
    async runLoop() {
        this.polling = true;
        let failureCount = 0;
        while (this.polling) {
            try {
                await this.pollIteration();
                if (failureCount > 0) {
                    (0, exports.log)(`Poll iteration recovered after ${failureCount} failures`);
                    failureCount = 0;
                }
            }
            catch (e) {
                (0, exports.log)("Failed poll iteration", {
                    failureCount,
                    error: e,
                });
                failureCount++;
            }
            await new Promise((resolve) => setTimeout(resolve, this.retryAfter * 1000));
        }
    }
    async pollIteration() {
        if (!this.clusterId) {
            throw new Error("Failed to poll. Could not find clusterId");
        }
        const tools = this.tools.map((fn) => fn.name);
        const pollResult = await this.client.listJobs({
            params: {
                clusterId: this.clusterId,
            },
            query: {
                tools: tools.join(","),
                status: "pending",
                acknowledge: true,
                limit: 10,
                waitTime: 20,
            },
        });
        const retryAfterHeader = pollResult.headers.get("retry-after");
        if (retryAfterHeader && !isNaN(Number(retryAfterHeader))) {
            this.retryAfter = Number(retryAfterHeader);
        }
        if (pollResult?.status === 410) {
            await (0, exports.registerMachine)(this.client, this.tools);
        }
        if (pollResult?.status !== 200) {
            throw new errors_1.InferableError("Failed to fetch calls", {
                status: pollResult?.status,
                body: pollResult?.body,
            });
        }
        const results = await Promise.allSettled(pollResult.body.map(async (job) => {
            await this.processCall(job);
        }));
        if (results.length > 0) {
            (0, exports.log)("Completed poll iteration", {
                results: results.map((r) => r.status),
            });
        }
    }
    async processCall(call) {
        const registration = this.tools.find((fn) => fn.name === call.function);
        if (!registration) {
            (0, exports.log)("Received call for unknown function", {
                function: call.function,
            });
            return;
        }
        (0, exports.log)("Executing job", {
            id: call.id,
            function: call.function,
            registered: !!registration,
        });
        const onComplete = async (result) => {
            (0, exports.log)("Persisting job result", {
                id: call.id,
                function: call.function,
                resultType: result.type,
                functionExecutionTime: result.functionExecutionTime,
            });
            await this.client
                .createJobResult({
                body: {
                    result: result.content,
                    resultType: result.type,
                    meta: {
                        functionExecutionTime: result.functionExecutionTime,
                    },
                },
                params: {
                    jobId: call.id,
                    clusterId: this.clusterId,
                },
            })
                .then(async (res) => {
                if (res.status === 204) {
                    (0, exports.log)("Completed job", call.id, call.function);
                }
                else {
                    throw new errors_1.InferableError(`Failed to persist call: ${res.status}`, {
                        jobId: call.id,
                        body: JSON.stringify(res.body),
                    });
                }
            });
        };
        const args = call.input;
        (0, exports.log)("Executing fn", {
            id: call.id,
            function: call.function,
            registeredFn: registration.func,
            args,
        });
        if (typeof args !== "object" || Array.isArray(args) || args === null) {
            (0, exports.log)("Function was called with invalid invalid format. Expected an object.", {
                function: call.function,
            });
            return onComplete({
                type: "rejection",
                content: (0, serialize_error_1.serializeError)(new Error("Function was called with invalid invalid format. Expected an object.")),
                functionExecutionTime: 0,
            });
        }
        try {
            (0, util_1.validateFunctionArgs)(registration.schema, args);
        }
        catch (e) {
            if (e instanceof zod_1.z.ZodError) {
                e.errors.forEach((error) => {
                    (0, exports.log)("Function input does not match schema", {
                        function: call.function,
                        path: error.path,
                        error: error.message,
                    });
                });
            }
            return onComplete({
                type: "rejection",
                content: (0, serialize_error_1.serializeError)(e),
                functionExecutionTime: 0,
            });
        }
        const result = await (0, execute_fn_1.executeFn)(registration.func, [
            args,
            {
                authContext: call.authContext,
                runContext: call.runContext,
                approved: call.approved,
            },
        ]);
        await onComplete(result);
    }
}
exports.PollingAgent = PollingAgent;
const registerMachine = async (client, tools) => {
    (0, exports.log)("registering machine", {
        tools: tools?.map((f) => f.name),
    });
    const registerResult = await client.createMachine({
        body: {
            tools: tools?.map((func) => ({
                name: func.name,
                description: func.description,
                schema: JSON.stringify((0, util_1.isZodType)(func.schema?.input)
                    ? (0, zod_to_json_schema_1.default)(func.schema?.input)
                    : func.schema?.input),
                config: func.config,
            })),
        },
    });
    if (registerResult?.status !== 200) {
        (0, exports.log)("Failed to register machine", registerResult);
        throw new errors_1.InferableAPIError("Failed to register machine", registerResult);
    }
    return {
        clusterId: registerResult.body.clusterId,
    };
};
exports.registerMachine = registerMachine;
//# sourceMappingURL=polling.js.map