import { z } from "zod";
import { ToolConfigSchema } from "./contract";
/**
 * Context object which is passed to function calls
 */
export type JobContext = {
    authContext?: unknown;
    runContext?: unknown;
    approved: boolean;
};
export type ToolConfig = z.infer<typeof ToolConfigSchema>;
export type ToolInput<T extends z.ZodTypeAny | JsonSchemaInput> = T extends z.ZodObject<infer Input> ? {
    [K in keyof Input]: z.infer<Input[K]>;
} : any;
/**
 * Schema for onStatusChange functions
 *
 * @see {@link https://docs.inferable.ai/pages/runs#onstatuschange}
 * @example
 * ```ts
 * inferable.default.register({
 *   name: "onStatusChangeFn",
 *   schema: {
 *    input: onStatusChangeInput
 *   },
 *   func: (_input) => {},
 * });
 * ```
 */
export declare const onStatusChangeInput: z.ZodObject<{
    runId: z.ZodString;
    status: z.ZodEnum<["pending", "running", "paused", "done", "failed"]>;
    result: z.ZodOptional<z.ZodNullable<z.ZodObject<{}, "passthrough", z.ZodTypeAny, z.objectOutputType<{}, z.ZodTypeAny, "passthrough">, z.objectInputType<{}, z.ZodTypeAny, "passthrough">>>>;
    summary: z.ZodOptional<z.ZodNullable<z.ZodString>>;
    tags: z.ZodOptional<z.ZodNullable<z.ZodRecord<z.ZodString, z.ZodString>>>;
}, "strip", z.ZodTypeAny, {
    status: "pending" | "running" | "paused" | "done" | "failed";
    runId: string;
    result?: z.objectOutputType<{}, z.ZodTypeAny, "passthrough"> | null | undefined;
    summary?: string | null | undefined;
    tags?: Record<string, string> | null | undefined;
}, {
    status: "pending" | "running" | "paused" | "done" | "failed";
    runId: string;
    result?: z.objectInputType<{}, z.ZodTypeAny, "passthrough"> | null | undefined;
    summary?: string | null | undefined;
    tags?: Record<string, string> | null | undefined;
}>;
/**
 * Schema for handleCustomAuth functions
 *
 * @see {@link https://docs.inferable.ai/pages/custom-auth}
 * @example
 * ```ts
 * inferable.default.register({
 *   name: "handleCustomAuth",
 *   schema: {
 *    input: handleCustomAuthInput
 *   },
 *   func: (_input) => {},
 * });
 * ```
 */
export declare const handleCustomAuthInput: z.ZodObject<{
    token: z.ZodString;
}, "strip", z.ZodTypeAny, {
    token: string;
}, {
    token: string;
}>;
import type { JSONSchema4Type } from "json-schema";
import type { JsonSchema7Type } from "zod-to-json-schema";
export type JsonSchema = JSONSchema4Type | JsonSchema7Type;
export type JsonSchemaInput = {
    type: string;
    properties: Record<string, JsonSchema>;
    required: string[];
    $schema: string;
};
export type ToolSchema<T extends z.ZodTypeAny | JsonSchemaInput> = {
    input: T;
};
export type ToolRegistrationInput<T extends z.ZodTypeAny | JsonSchemaInput> = {
    name: string;
    func: (input: ToolInput<T>, context: JobContext) => Promise<any>;
    schema?: ToolSchema<T>;
    config?: ToolConfig;
    description?: string;
};
export type WorkflowToolRegistrationInput<T extends z.ZodTypeAny | JsonSchemaInput> = {
    name: string;
    inputSchema?: T;
    config?: ToolConfig;
    func: (input: ToolInput<T>, context: JobContext) => Promise<unknown>;
};
