"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Workflow = exports.helpers = void 0;
const zod_1 = require("zod");
const zod_to_json_schema_1 = __importDefault(require("zod-to-json-schema"));
const cybr53_1 = require("../util/cybr53");
const errors_1 = require("../errors");
const polling_1 = require("../polling");
const util_1 = require("../util");
const l1m_1 = __importDefault(require("l1m"));
class WorkflowPausableError extends Error {
    constructor(message) {
        super(message);
        this.name = "WorkflowPausableError";
    }
}
class WorkflowTerminableError extends Error {
    constructor(message) {
        super(message);
        this.name = "WorkflowTerminableError";
    }
}
exports.helpers = {
    structuredPrompt: (params) => {
        return [
            "# Facts",
            ...params.facts.map((f) => `- ${f}`),
            "# Your goals",
            ...params.goals.map((g) => `- GOAL: ${g}`),
        ].join("\n");
    },
};
class Workflow {
    constructor(config) {
        this.versionHandlers = new Map();
        this.agentTools = [];
        this.tools = {
            register: (tool) => {
                this.logger?.info("Registering tool", {
                    name: this.name,
                    tool,
                });
                this.agentTools.push(tool);
            },
        };
        this.name = config.name;
        this.description = config.description;
        this.inputSchema = config.inputSchema;
        this.logger = config.logger;
        this.client = config.client;
        this.getClusterId = config.getClusterId;
        this.endpoint = config.endpoint;
        this.machineId = config.machineId;
        this.apiSecret = config.apiSecret;
        this.config = config.config;
    }
    version(version) {
        return {
            define: (handler) => {
                this.logger?.info("Defining workflow handler", {
                    version,
                    name: this.name,
                });
                this.versionHandlers.set(version, handler);
            },
        };
    }
    createWorkflowContext(version, executionId, input, jobCtx, clusterId) {
        this.logger?.info("Creating workflow context", {
            version,
            name: this.name,
            executionId,
        });
        const l1m = new l1m_1.default({
            baseUrl: `${this.endpoint}/clusters/${clusterId}/l1m`,
            additionalHeaders: {
                "x-workflow-execution-id": executionId,
                Authorization: `Bearer ${this.apiSecret}`,
            },
            provider: {
                model: "claude-3-5-sonnet",
                key: "",
                url: "",
            },
        });
        const memo = async (name, fn) => {
            const ctx = this.createWorkflowContext(version, executionId, input, jobCtx, clusterId);
            const serialize = (value) => JSON.stringify({ value });
            const deserialize = (value) => {
                try {
                    return JSON.parse(value).value;
                }
                catch {
                    return null;
                }
            };
            const existingValue = await this.client.getClusterKV({
                params: {
                    clusterId: await this.getClusterId(),
                    key: `${executionId}_memo_${name}`,
                },
            });
            if (existingValue.status === 200) {
                const existingValueParsed = deserialize(existingValue.body.value);
                if (existingValueParsed) {
                    return existingValueParsed;
                }
            }
            const result = await fn(ctx);
            // TODO: async/retry
            const setResult = await this.client.setClusterKV({
                params: {
                    clusterId: await this.getClusterId(),
                    key: `${executionId}_memo_${name}`,
                },
                body: {
                    value: serialize(result),
                    onConflict: "doNothing",
                },
            });
            if (setResult.status !== 200) {
                this.logger?.error("Failed to set result", {
                    name,
                    executionId,
                    status: setResult.status,
                });
                throw new Error("Failed to set result");
            }
            return deserialize(setResult.body.value);
        };
        const agents = {
            react: async (config) => {
                this.logger?.info("Running agent in workflow", {
                    version,
                    name: this.name,
                    executionId,
                    agentName: config.name,
                });
                const resultSchema = config.schema
                    ? (0, zod_to_json_schema_1.default)(config.schema)
                    : undefined;
                const runId = `${executionId}_${config.name}_${(0, cybr53_1.cyrb53)(JSON.stringify([
                    config.instructions,
                    executionId,
                    resultSchema,
                    this.name,
                    version,
                    config.input,
                ].join(".")))}`;
                const result = await this.client.createRun({
                    params: {
                        clusterId: await this.getClusterId(),
                    },
                    headers: {
                        "x-provider-key": config.provider?.key,
                        "x-provider-model": config.provider?.model,
                    },
                    body: {
                        name: `${this.name}_${config.name}`,
                        id: runId,
                        systemPrompt: config.instructions,
                        resultSchema,
                        tools: config.tools.map((t) => `tool_${this.name}_${t}`),
                        onStatusChange: {
                            type: "workflow",
                            statuses: ["failed", "done"],
                            workflow: {
                                executionId: executionId,
                            },
                        },
                        tags: {
                            "workflow.name": this.name,
                            "workflow.version": version.toString(),
                            "workflow.executionId": executionId,
                        },
                        initialPrompt: config.input,
                        interactive: true,
                    },
                });
                this.logger?.info("Agent run created", {
                    version,
                    name: this.name,
                    executionId,
                    agentName: config.name,
                    runId,
                    status: result.status,
                    result: result.body,
                });
                if (result.status !== 201) {
                    // TODO: Add better error handling
                    throw new errors_1.InferableAPIError(`Failed to create run: ${result.status}`, result);
                }
                if (result.body.status === "done") {
                    if (!config.onBeforeReturn) {
                        return result.body.result;
                    }
                    let shouldReturn = true;
                    await config.onBeforeReturn(result.body.result, {
                        sendMessage: async (message) => {
                            await this.client.createMessage({
                                params: {
                                    clusterId: await this.getClusterId(),
                                    runId,
                                },
                                body: {
                                    message,
                                    type: "human",
                                },
                            });
                            shouldReturn = false;
                        },
                    });
                    if (shouldReturn) {
                        return result.body.result;
                    }
                    else {
                        throw new WorkflowPausableError(`Agent ${config.name} is not done.`);
                    }
                }
                else if (result.body.status === "failed") {
                    throw new WorkflowTerminableError(`Agent ${config.name} failed. As a result, we've failed the entire workflow (executionId: ${executionId}). Please refer to run failure details for more information.`);
                }
                else {
                    throw new WorkflowPausableError(`Agent ${config.name} is not done.`);
                }
            },
        };
        return {
            ...jobCtx,
            llm: l1m,
            result: memo,
            memo,
            agents,
            /**
             * @deprecated Use `agents` instead
             */
            agent: (config) => {
                return {
                    trigger: async (params) => {
                        this.logger?.info("Running agent in workflow", {
                            version,
                            name: this.name,
                            executionId,
                            agentName: config.name,
                        });
                        const resultSchema = config.resultSchema
                            ? (0, zod_to_json_schema_1.default)(config.resultSchema)
                            : undefined;
                        const runId = config.runId
                            ? `${executionId}_${config.name}_${config.runId}`
                            : `${executionId}_${config.name}_${(0, cybr53_1.cyrb53)(JSON.stringify([
                                config.systemPrompt,
                                executionId,
                                resultSchema,
                                this.name,
                                version,
                                params.data,
                            ]))}`;
                        const result = await this.client.createRun({
                            params: {
                                clusterId: await this.getClusterId(),
                            },
                            body: {
                                name: `${this.name}_${config.name}`,
                                id: runId,
                                systemPrompt: config.systemPrompt,
                                resultSchema,
                                tools: config.tools,
                                onStatusChange: {
                                    type: "workflow",
                                    statuses: ["failed", "done"],
                                    workflow: {
                                        executionId: executionId,
                                    },
                                },
                                tags: {
                                    "workflow.name": this.name,
                                    "workflow.version": version.toString(),
                                    "workflow.executionId": executionId,
                                },
                                initialPrompt: JSON.stringify(params.data),
                                interactive: false,
                            },
                        });
                        this.logger?.info("Agent run created", {
                            version,
                            name: this.name,
                            executionId,
                            agentName: config.name,
                            runId,
                            status: result.status,
                            result: result.body,
                        });
                        if (result.status !== 201) {
                            // TODO: Add better error handling
                            throw new errors_1.InferableAPIError(`Failed to create run: ${result.status}`, result);
                        }
                        if (result.body.status === "done") {
                            return {
                                result: result.body.result,
                            };
                        }
                        else if (result.body.status === "failed") {
                            throw new WorkflowTerminableError(`Run ${runId} failed. As a result, we've failed the entire workflow (executionId: ${executionId}). Please refer to run failure details for more information.`);
                        }
                        else {
                            throw new WorkflowPausableError(`Run ${runId} is not done.`);
                        }
                    },
                };
            },
            input,
            log: async (status, meta) => {
                await this.client.createWorkflowLog({
                    params: {
                        clusterId: await this.getClusterId(),
                        workflowName: this.name,
                        executionId,
                    },
                    body: { status, data: meta },
                });
            },
            notify: async (notification) => {
                await this.client.createWorkflowNotification({
                    params: {
                        clusterId: await this.getClusterId(),
                        workflowName: this.name,
                        executionId,
                    },
                    body: notification,
                });
            },
        };
    }
    async listen() {
        if (this.pollingAgent) {
            throw new errors_1.InferableError("Workflow already listening");
        }
        this.logger?.info("Starting workflow listeners", {
            name: this.name,
            versions: Array.from(this.versionHandlers.keys()),
        });
        const tools = this.agentTools.map((tool) => ({
            name: `tool_${this.name}_${tool.name}`,
            func: tool.func,
            schema: {
                input: tool.inputSchema ?? zod_1.z.object({}),
            },
            config: tool.config,
        }));
        this.versionHandlers.forEach((handler, version) => {
            tools.push({
                func: async (input, jobCtx) => {
                    if (!input.executionId) {
                        const error = "executionId field must be provided in the workflow input, and must be a string";
                        this.logger?.error(error);
                        throw new Error(error);
                    }
                    const ctx = this.createWorkflowContext(version, input.executionId, input, jobCtx, await this.getClusterId());
                    try {
                        return await handler(ctx, input);
                    }
                    catch (e) {
                        if (e instanceof WorkflowPausableError) {
                            return util_1.Interrupt.general();
                        }
                        throw e;
                    }
                },
                name: `workflows_${this.name}_${version}`,
                description: this.description,
                schema: {
                    input: this.inputSchema,
                },
                config: {
                    ...this.config,
                    private: true,
                },
            });
        });
        this.pollingAgent = new polling_1.PollingAgent({
            endpoint: this.endpoint,
            machineId: this.machineId,
            apiSecret: this.apiSecret,
            clusterId: await this.getClusterId(),
            tools,
        });
        try {
            await this.pollingAgent.start();
            this.logger?.info("Workflow listeners started", { name: this.name });
        }
        catch (error) {
            this.logger?.error("Failed to start workflow listeners", {
                name: this.name,
                error,
            });
            throw error;
        }
    }
    async unlisten() {
        this.logger?.info("Stopping workflow listeners", { name: this.name });
        try {
            await this.pollingAgent?.stop();
            this.logger?.info("Workflow listeners stopped", { name: this.name });
        }
        catch (error) {
            this.logger?.error("Failed to stop workflow listeners", {
                name: this.name,
                error,
            });
            throw error;
        }
    }
}
exports.Workflow = Workflow;
//# sourceMappingURL=workflow.js.map